# -*- coding:utf-8 -*-

import six
import functools


__all__ = ['cache']


if six.PY3:
    cache = functools.lru_cache
else:
    def cache(func):
        _cache = func.cache = {}

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return _cache[args]
            except KeyError:
                rv = func(*args, **kwargs)
                _cache[args] = rv
                return rv
        return wrapper



# -*- coding:utf-8 -*-

import six
import numpy as np
from collections import deque

__all__ = ['is_integer', 'is_float', 'is_seq']



def is_integer(value):
    return isinstance(value, six.integer_types + (np.integer,))


def is_float(value):
    return isinstance(value, (float, np.floating))


def is_seq(value):
    if isinstance(value, (list, tuple, deque)):
        return True
    elif isinstance(value, np.ndarray):
        return value.ndim == 1

    return False



# -*- coding:utf-8 -*-

import six
import sys
import re

__all__ = ['str2list', 'dict2str', 'Str2Obj', 'str2obj']


def str2list(s, pattern=',|;|:|#|&|\||\s+'):
    l = re.split(pattern, s)
    if len(l) == 1 and l[0] == '':
        return []
    else:
        return l


def dict2str(d, fmt='{}={}', delimiter='&', sort=True):
    keys = sorted(d.keys()) if sort else d.keys()
    tokens = [fmt.format(k, d[k]) for k in keys]
    return delimiter.join(tokens)


class Str2Obj(object):
    """Str2Obj converts str to actual object.
    inspired by django's ContentTyper
    """
    def __init__(self):
        self.cache = {}
        self.registered = {}

    def register(self, name, func):
        """
        register a generator func.
        :param name: generator name
        :param func: generator func
        :return: info dict.
        """
        self.registered[name] = func

    def deregister(self, name=None):
        if name is None:
            self.registered.clear()
        else:
            self.registered.pop(name, None)

    def clear_cache(self, name=None):
        if name is None:
            self.cache.clear()
        else:
            self.cache.pop(name, None)

    def _return_or_raise(self, value, msg=None):
        if isinstance(value, six.class_types) and issubclass(value, Exception):
            raise value(msg)
        elif isinstance(value, Exception):
            raise value
        else:
            return value

    def __call__(self, s, entrance=None, default=ValueError, cache=True):
        """

        :param s: the string to be converted.
            which can be like 'a', 'a.b.c', 'f()', 'f("arg", kw="v", ...)';
            if s is not a string, it will be returned as it is.
        :param entrance: if provided, retrieve the object from this dict.
        :param default: a default value or an exception class/object to be raised. default: ValueError;
        :param cache: whether to use cache.
        :return: converted object.
        """
        if not isinstance(s, six.string_types):
            return s

        m = re.match(r'^(?P<name>.*?)(\((?P<args>.*)\))?$', s)
        name = m.group('name')
        if name in self.registered:
            obj = self.registered[name]
        else:
            if entrance is None:
                import importlib

                if name in self.cache:
                    obj = self.cache[name]
                else:
                    tokens = name.split('.')
                    try:
                        if len(tokens) > 1:
                            obj = importlib.import_module(tokens[0])
                            for i in range(1, len(tokens)):
                                try:
                                    obj = getattr(obj, tokens[i])
                                except AttributeError:
                                    obj = importlib.import_module('.'.join(tokens[:i+1]))
                        else:
                            scope = sys.modules['__main__'].__dict__
                            obj = scope[tokens[0]]
                        if cache:
                            self.cache[s] = obj
                    except Exception as e:
                        # self.cache[s] = Null
                        return self._return_or_raise(default, repr(e))

            else:
                # not using cache
                tokens = name.split('.')
                obj = entrance
                try:
                    for tk in tokens:
                        try:
                            obj = obj[tk]
                        except TypeError:
                            try:
                                obj = getattr(obj, tk)
                            except AttributeError:
                                raise ValueError(u"Cannot get %s in %s" % (name, entrance))
                except Exception as e:
                    return self._return_or_raise(default, msg=repr(e))

        raw_args = m.group('args')
        if raw_args is None:
            return obj
        else:
            try:
                res = eval('obj({})'.format(raw_args))
                return res
            except Exception as e:
                return self._return_or_raise(default, repr(e))


str2obj = Str2Obj()



# -*- coding:utf-8 -*-

import six
import re


__all__ = ['simple_pattern_to_regex', 'grep']


def simple_pattern_to_regex(pattern, sep_chars=None, keep_dots=False, strict=False):
    """simple_pattern_to_regex converts a simple pattern into a regex pattern.
    :param pattern: a curly-bracketed pattern, like str's named format pattern, e.g. '{dataset}/{varname}.csv'
    :param sep_chars: chars that is always considered as delimiters, e.g., "/" in file path. Default: None,
    :param keep_dots: if False, "." is converted to "\.". Default: False.
    :param strict: if True, add "^" and "$" to the begining and end of the pattern. Default: False
    :return: regex pattern, e.g. r"(?P<dataset>[^/]*)/(?P<varname>[^/]*)\.nc"
    """
    try:
        valid_char_pattern = r'[^{}]'.format(sep_chars) if sep_chars else r'.'
        re_pattern = re.sub(r'\{([a-zA-Z_][^:}]*)(:[^}]*)*\}', r'(?P<\1>{}*)'.format(valid_char_pattern), pattern)
        if re_pattern != pattern:
            if not keep_dots:
                if valid_char_pattern == r'.':
                    # this is an ugly fix, to make sure that '.' is replaced only in the raw pattern.
                    re_pattern = re.sub(r'\.', '\.', pattern)
                    re_pattern = re.sub(r'\{([a-zA-Z_][^:}]*)(:[^}]*)*\}', r'(?P<\1>{}*)'.format(valid_char_pattern), re_pattern)
                else:
                    re_pattern = re.sub(r'\.', '\.', re_pattern)

            if strict:
                re_pattern = '^' + re_pattern.lstrip('^')
                re_pattern = re_pattern.rstrip('$') + '$'
            return re_pattern
        else:
            return pattern
    except Exception as e:
        return pattern


def grep(pattern, seq, flags=0, parse_info=False):
    """grep greps patterns from seqs.
Parameters:
    pattern: a regex pattern (str or compiled), a simple pattern, or a function-like object that returns True/False or info dict.
    seq: seq of str or any other stuffs, etc. If pattern is not func and seq[0] is not string type, seq will be converted using repr() first.
        items in seq can also be 2-tuple: (object-itself, string-representation), pattern is applied on the latter.
    flags: 'ILMSUX' or number of re.I|re.L|re.M|re.S|re.U|re.X .
        I: ignore case;
        L: locale dependent;
        M: multi-line;
        S: dot matches all;
        U: unicode dependent;
        X: verbose.
    parse_info: whether to return parsed info (e.g., group info of regex, or the result of func-like pattern)
Returns:
    A filtered list of matched str or (matched_str, groupdict). The latter applies only when the pattern has groups.
    """
    if isinstance(pattern, six.string_types) and '(?P<' not in pattern:
        pattern = simple_pattern_to_regex(pattern)

    if isinstance(flags, six.string_types):
        true_flags = 0
        flags = flags.upper()
        for letter in flags:
            if letter in 'ILMSUX':
                true_flags += re.__dict__[letter]
    else:
        true_flags = flags

    res = []
    if isinstance(pattern, six.string_types + (type(re.compile(r'')),)):
        for item in seq:
            if isinstance(item, (tuple, list)) and len(item) == 2:
                object, grep_target = item
            else:
                object = item
                grep_target = item if isinstance(item, six.string_types) else repr(item)
            m = re.search(pattern, grep_target, true_flags)
            if not m:
                continue

            info = m.groupdict().copy()
            groups = m.groups()
            info.update(dict(zip(range(1, len(groups)+1), groups)))
            if parse_info:
                res.append((object, info))
            else:
                res.append(object)
    elif callable(pattern):
        for item in seq:
            if isinstance(item, tuple) and len(item) == 2:
                object, grep_target = item
            else:
                object = item
                grep_target = item if isinstance(item, six.string_types) else repr(item)

            info = pattern(grep_target)
            if info:
                if parse_info:
                    if not isinstance(info, dict):
                        res.append((object, {}))
                    else:
                        res.append((object, info))
                else:
                    res.append(object)

    return res



# -*- coding:utf-8 -*-

import six
import unittest

import os
import sys
import re
import datetime

from krux.converters import *


class TestStr2Obj(unittest.TestCase):
    def test_import(self):
        for s, obj in [
            ('sys.path', sys.path),
            ('datetime.datetime.now', datetime.datetime.now),

        ]:
            self.assertEqual(str2obj(s), obj)

    def test_raise(self):
        with self.assertRaises(ValueError):
            str2obj('not_a_lib.def.g')

        with self.assertRaises(RuntimeError):
            str2obj('not_a_lib.def.g', default=RuntimeError)

    def test_default(self):
        self.assertEqual(str2obj('not_a_lib.def.g', default='foo'), 'foo')

    def test_dict(self):
        class C(object):
            x = 5

        d = {
            "a": {
                "b": 1
            },
            "c": C()
        }

        default = 99
        for s, val in [
            ('a.b', 1),
            ('c.x', 5),
            ('c.y', default),
            ('d', default),
            ('a.b.c', default),
        ]:
            self.assertEqual(str2obj(s, entrance=d, default=default), val)

    def test_func(self):
        self.assertEqual(str2obj('operator.add(3, 5)'), 8)


if __name__ == '__main__':
    unittest.main()


# -*- coding:utf-8 -*-


__all__ = ['list2dict', 'dict2list']


def list2dict(l):
    if not l:
        return {}

    keys = l[0].keys()
    d = {}
    for k in keys:
        d[k] = [item[k] for item in l]

    return d


def dict2list(d):
    if not d:
        return []

    keys = d.keys()
    zipped = zip(*d.values())
    return [dict(zip(keys, tup)) for tup in zipped]